package com.trafficparrot.sdk.jms;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.github.jknack.handlebars.Handlebars;
import com.github.jknack.handlebars.HandlebarsException;
import com.github.jknack.handlebars.Template;
import com.google.common.annotations.VisibleForTesting;
import com.google.common.base.Optional;

import javax.jms.BytesMessage;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.TextMessage;
import java.io.IOException;
import java.util.Map;

import static com.trafficparrot.sdk.handlebars.HandlebarsFactory.handlebars;
import static com.trafficparrot.sdk.jms.BytesMessageReadBytes.readBytes;
import static com.trafficparrot.sdk.messaging.ObjectJsonRepresentation.objectToJson;
import static com.trafficparrot.sdk.messaging.ObjectSerializedRepresentation.OBJECT_SERIALIZED_REPRESENTATION;
import static com.trafficparrot.sdk.messaging.PrintableCharactersRepresentation.encodePrintableCharacters;
import static java.lang.String.format;
import static java.util.Collections.emptyMap;
import static java.util.Collections.singletonMap;

public class JmsHandlebarsResponseTransformer {

    private final Handlebars handlebars;
    private final boolean enabled;
    private final TpLogger logger;

    public JmsHandlebarsResponseTransformer(Handlebars handlebars, boolean enabled, TpLogger logger) {
        this.handlebars = handlebars;
        this.enabled = enabled;
        this.logger = logger;
    }

    public static JmsHandlebarsResponseTransformer handlebarsTransformer(String commaSeparatedHandlebarHelpers, Boolean handlebarsEnabled, TpLogger logger) {
        if (handlebarsEnabled) {
            return defineHelpers(logger, handlebars(commaSeparatedHandlebarHelpers));
        }
        return handlebarsAreDisabled(logger);
    }

    public static JmsHandlebarsResponseTransformer handlebarsAreDisabled(TpLogger logger) {
        return new JmsHandlebarsResponseTransformer(null, false, logger);
    }

    private static JmsHandlebarsResponseTransformer defineHelpers(TpLogger logger, Handlebars handlebars) {
        return new JmsHandlebarsResponseTransformer(handlebars, true, logger);
    }

    public JmsResponse transform(Message requestMessage, JmsResponse response) {
        if (enabled) {
            try {
                Template responseTemplate = handlebars.compileInline(response.text);
                String responseTextWithTemplateApplied = responseTemplate.apply(requestModel(requestMessage));
                return new JmsResponse(response.destination, response.jmsResponseTransformerClassName, responseTextWithTemplateApplied, response.bodyType);
            } catch (HandlebarsException | IOException | JMSException e) {
                String errorMessage = format("Problem while applying handlebars: \n\n%s\n\nOriginal text: \n%s", e.getMessage(), response.text);
                logger.error(errorMessage, e);
                return new JmsResponse(response.destination, errorMessage + ". \n\nSee logs for more details.");
            }
        } else {
            return response;
        }
    }

    @VisibleForTesting
    protected Map<String, ?> requestModel(Message requestMessage) throws JMSException {
        if (requestMessage instanceof TextMessage) {
            TextMessage textMessage = (TextMessage) requestMessage;
            return requestBody(textMessage.getText());
        } else if (requestMessage instanceof BytesMessage) {
            BytesMessage bytesMessage = (BytesMessage) requestMessage;
            byte[] bytes = readBytes(bytesMessage);
            Optional<Object> object = OBJECT_SERIALIZED_REPRESENTATION.tryToReadObject(bytes);
            if (!object.isPresent()) {
                return bytesMessageBody(bytes);
            }
            try {
                return requestBody(objectToJson(object.get()));
            } catch (JsonProcessingException e) {
                logger.error(format("Captured request object '%s' will not be available in the templating model because Traffic Parrot cannot store it in a JSON intermediate representation.", object.get()), e);
                return bytesMessageBody(bytes);
            }
        } else {
            return emptyMap();
        }
    }

    private Map<String, Map<String, String>> bytesMessageBody(byte[] bytes) {
        return requestBody(encodePrintableCharacters(bytes));
    }

    private Map<String, Map<String, String>> requestBody(String text) {
        return singletonMap("request", singletonMap("body", text));
    }
}
