package com.trafficparrot.sdk.jms;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.google.common.collect.ComparisonChain;
import com.trafficparrot.sdk.messaging.Base64Representation;

import javax.jms.BytesMessage;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.Session;
import java.io.IOException;

import static com.trafficparrot.sdk.jms.JmsMessageType.jmsMessageType;
import static com.trafficparrot.sdk.jms.MessageBodyType.*;
import static com.trafficparrot.sdk.messaging.Base64Representation.decodeBase64;
import static com.trafficparrot.sdk.messaging.ObjectJsonRepresentation.jsonToObject;
import static com.trafficparrot.sdk.messaging.ObjectJsonRepresentation.objectToJson;
import static com.trafficparrot.sdk.messaging.ObjectSerializedRepresentation.OBJECT_SERIALIZED_REPRESENTATION;
import static com.trafficparrot.sdk.messaging.PrintableCharactersRepresentation.decodePrintableCharacters;
import static com.trafficparrot.sdk.messaging.PrintableCharactersRepresentation.encodePrintableCharacters;
import static org.apache.commons.lang3.ObjectUtils.firstNonNull;
import static org.apache.commons.lang3.builder.ToStringBuilder.reflectionToString;
import static org.apache.commons.lang3.builder.ToStringStyle.SHORT_PREFIX_STYLE;

public class JmsResponse implements Comparable<JmsResponse> {
    public static final String NO_TRANSFORMER = "NO_TRANSFORMER";

    public final Destination destination;
    public final String jmsResponseTransformerClassName;
    public final String text;
    public final MessageBodyType bodyType;
    public final String jmsMessageType;

    public JmsResponse(@JsonProperty("destination") Destination destination,
                       @JsonProperty("jmsResponseTransformerClassName") String jmsResponseTransformerClassName,
                       @JsonProperty("text") String text,
                       @JsonProperty("bodyType") MessageBodyType bodyType,
                       @JsonProperty("jmsMessageType") String jmsMessageType) {
        this.destination = destination;
        this.jmsResponseTransformerClassName = jmsResponseTransformerClassName;
        this.text = text;
        this.bodyType = firstNonNull(bodyType, TEXT);
        this.jmsMessageType = firstNonNull(jmsMessageType, jmsMessageType(this.bodyType));
    }

    public JmsResponse(Destination destination, String jmsResponseTransformerClassName, String text, MessageBodyType bodyType) {
        this(destination, jmsResponseTransformerClassName, text, bodyType, jmsMessageType(bodyType));
    }

    public JmsResponse(Destination destination, String jmsResponseTransformerClassName, String text) {
        this(destination, jmsResponseTransformerClassName, text, TEXT);
    }

    public JmsResponse(Destination destination, String text) {
        this(destination, NO_TRANSFORMER, text);
    }

    public JmsResponse(Destination destination, String jmsResponseTransformerClassName, Object object) throws JsonProcessingException {
        this(destination, jmsResponseTransformerClassName, objectToJson(object), javaClassType(object));
    }

    public JmsResponse(Destination destination, Object object) throws JsonProcessingException {
        this(destination, NO_TRANSFORMER, object);
    }

    public static JmsResponse base64BytesResponse(Destination destination, byte[] bytes) {
        return base64BytesResponse(destination, NO_TRANSFORMER, bytes);
    }

    public static JmsResponse base64BytesResponse(Destination destination, String jmsResponseTransformerClassName, byte[] bytes) {
        return new JmsResponse(destination, jmsResponseTransformerClassName, Base64Representation.encodeBase64(bytes), BYTES);
    }

    public static JmsResponse printableCharactersBytesResponse(Destination destination, byte[] bytes) {
        return printableCharactersBytesResponse(destination, NO_TRANSFORMER, bytes);
    }

    public static JmsResponse printableCharactersBytesResponse(Destination destination, String jmsResponseTransformerClassName, byte[] bytes) {
        return new JmsResponse(destination, jmsResponseTransformerClassName, encodePrintableCharacters(bytes), PRINTABLE_CHARACTERS);
    }

    @Override
    @JsonIgnore
    public String toString() {
        return reflectionToString(this, SHORT_PREFIX_STYLE);
    }

    @Override
    public int compareTo(JmsResponse other) {
        return ComparisonChain.start()
                .compare(destination, other.destination)
                .compare(text, other.text)
                .result();
    }

    public Message createResponseMessage(Session targetSession) throws JMSException {
        if (BYTES.equals(bodyType)) {
            BytesMessage bytesMessage = targetSession.createBytesMessage();
            bytesMessage.writeBytes(decodeBase64(text));
            return bytesMessage;
        } else if (PRINTABLE_CHARACTERS.equals(bodyType)) {
            BytesMessage bytesMessage = targetSession.createBytesMessage();
            bytesMessage.writeBytes(decodePrintableCharacters(text));
            return bytesMessage;
        } else if (TEXT.equals(bodyType)) {
            return targetSession.createTextMessage(text);
        } else if (bodyType.isJavaObject()) {
            return javaObjectMessage(targetSession);
        } else {
            throw new IllegalStateException("Unrecognised response type: " + bodyType);
        }
    }

    public void validate() throws IOException {
        if (bodyType.isJavaObject()) {
            String textWithoutTemplateTokens = text.replaceAll("\\{\\{.*}}", "null");
            jsonToObject(textWithoutTemplateTokens, bodyType.javaClass());
        }
    }

    private Message javaObjectMessage(Session targetSession) throws JMSException {
        try {
            Object object = jsonToObject(text, bodyType.javaClass());
            byte[] bytes = OBJECT_SERIALIZED_REPRESENTATION.writeObjectBytes(object);

            BytesMessage bytesMessage = targetSession.createBytesMessage();
            bytesMessage.writeBytes(bytes);

            return bytesMessage;
        } catch (IOException e) {
            throw new IllegalStateException("Problem creating java object message", e);
        }
    }
}
